"""Sad path tests for SPARQLOperationParametersConstructor."""

import pytest
from sparqlx.utils.operation_parameters import (
    QueryOperationParametersConstructor,
    UpdateOperationParametersConstructor,
)


def test_query_operation_parameters_invalid_method():
    msg = "Expected query method 'GET', 'POST' or 'POST-direct'. Got 'invalid'."
    with pytest.raises(ValueError, match=msg):
        QueryOperationParametersConstructor(
            query="select * where {}", query_type="SelectQuery"
        ).get_params("invalid")  # type: ignore


def test_update_operation_parameters_invalid_method():
    msg = "Expected update method 'POST' or 'POST-direct'. Got 'invalid'."
    with pytest.raises(ValueError, match=msg):
        UpdateOperationParametersConstructor(
            update_request="insert {} where {}"
        ).get_params("invalid")  # type: ignore



"""SPARQLx testing utils."""

import asyncio
from collections.abc import AsyncIterator, Iterable
from contextlib import AbstractContextManager, asynccontextmanager
from typing import Any
from urllib.parse import parse_qs

import httpx
from sparqlx.types import SPARQLResultBinding, SPARQLResultBindingValue


def parse_response_qs(response: httpx.Response) -> dict[str, list]:
    content = response.request.content.decode("utf-8")
    return parse_qs(content)


async def acall(obj: Any, method: str, *args, **kwargs):
    f = getattr(obj, method)

    return (
        await f(*args, **kwargs)
        if asyncio.iscoroutinefunction(f)
        else f(*args, **kwargs)
    )


def sparql_result_set_equal(
    result_1: Iterable[SPARQLResultBinding], result_2: Iterable[SPARQLResultBinding]
) -> bool:
    def freeze(
        result: Iterable[SPARQLResultBinding],
    ) -> set[frozenset[tuple[str, SPARQLResultBindingValue]]]:
        return {frozenset(binding.items()) for binding in result}

    return freeze(result_1) == freeze(result_2)


@asynccontextmanager
async def as_async_cm[T](sync_cm: AbstractContextManager[T]) -> AsyncIterator[T]:
    """Async context wrapper around a sync context manager.

    The async context manager allows to call a sync context manager
    from an async context statement. This workaround is mentioned in PEP 806,
    see https://peps.python.org/pep-0806/#workaround-an-as-acm-wrapper.
    """
    with sync_cm as result:
        await asyncio.sleep(0)
        yield result



import abc
from collections import UserDict
from collections.abc import Mapping
from typing import Literal as TLiteral, NamedTuple

from sparqlx.types import (
    RequestDataValue,
    SPARQLQuery,
    SPARQLQueryTypeLiteral,
    SPARQLResponseFormat,
)


class MimeTypeMap(UserDict):
    def __missing__(self, key):
        return key


sparql_result_response_format_map = MimeTypeMap(
    {
        "json": "application/sparql-results+json",
        "xml": "application/sparql-results+xml",
        "csv": "text/csv",
        "tsv": "text/tab-separated-values",
    }
)

rdf_response_format_map = MimeTypeMap(
    {
        "turtle": "text/turtle",
        "xml": "application/rdf+xml",
        "ntriples": "application/n-triples",
        "json-ld": "application/ld+json",
    }
)


class SPARQLOperationDataMap(UserDict):
    def __init__(self, **kwargs):
        self.data = {k.replace("_", "-"): v for k, v in kwargs.items() if v is not None}


class SPARQLOperationParameters(NamedTuple):
    method: TLiteral["GET", "POST"]
    headers: Mapping
    content: str | None = None
    data: Mapping | None = None
    params: Mapping | None = None


class SPARQLOperationParametersConstructor(abc.ABC):
    @abc.abstractmethod
    def get_params(self, method) -> SPARQLOperationParameters: ...


class QueryOperationParametersConstructor(SPARQLOperationParametersConstructor):
    def __init__(
        self,
        query: SPARQLQuery,
        query_type: SPARQLQueryTypeLiteral,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
    ) -> None:
        self._query = query
        self._query_type = query_type
        self._response_format = response_format
        self._version = version
        self._default_graph_uri = default_graph_uri
        self._named_graph_uri = named_graph_uri

    def get_params(
        self, method: TLiteral["GET", "POST", "POST-direct"]
    ) -> SPARQLOperationParameters:
        match method:
            case "GET":
                return self._build_get_params()
            case "POST":
                return self._build_post_params()
            case "POST-direct":
                return self._build_post_direct_params()
            case _:
                msg = (
                    "Expected query method 'GET', 'POST' or 'POST-direct'. "
                    f"Got '{method}'."
                )
                raise ValueError(msg)

    def _build_get_params(self) -> SPARQLOperationParameters:
        return SPARQLOperationParameters(
            method="GET",
            headers={"Accept": self._get_response_format()},
            params=SPARQLOperationDataMap(
                query=self._query,
                version=self._version,
                default_graph_uri=self._default_graph_uri,
                named_graph_uri=self._named_graph_uri,
            ),
        )

    def _build_post_params(self) -> SPARQLOperationParameters:
        return SPARQLOperationParameters(
            method="POST",
            headers={
                "Accept": self._get_response_format(),
                "Content-Type": "application/x-www-form-urlencoded",
            },
            data=SPARQLOperationDataMap(
                query=self._query,
                version=self._version,
                default_graph_uri=self._default_graph_uri,
                named_graph_uri=self._named_graph_uri,
            ),
        )

    def _build_post_direct_params(self) -> SPARQLOperationParameters:
        return SPARQLOperationParameters(
            method="POST",
            headers={
                "Accept": self._get_response_format(),
                "Content-Type": (
                    "application/sparql-query"
                    f"{'' if self._version is None else f'; version={self._version}'}"
                ),
            },
            content=self._query,
            params=SPARQLOperationDataMap(
                default_graph_uri=self._default_graph_uri,
                named_graph_uri=self._named_graph_uri,
            ),
        )

    def _get_response_format(self) -> str:
        match self._query_type:
            case "SelectQuery" | "AskQuery":
                _response_format = sparql_result_response_format_map[
                    self._response_format or "json"
                ]
            case "DescribeQuery" | "ConstructQuery":
                _response_format = rdf_response_format_map[
                    self._response_format or "turtle"
                ]
            case _:  # pragma: no cover
                raise ValueError(f"Unsupported query type: {self._query_type}")

        return _response_format


class UpdateOperationParametersConstructor(SPARQLOperationParametersConstructor):
    def __init__(
        self,
        update_request: str,
        version: str | None = None,
        using_graph_uri: RequestDataValue = None,
        using_named_graph_uri: RequestDataValue = None,
    ):
        self._update_request = update_request
        self._version = version
        self._using_graph_uri = using_graph_uri
        self._using_named_graph_uri = using_named_graph_uri

    def get_params(
        self, method: TLiteral["POST", "POST-direct"]
    ) -> SPARQLOperationParameters:
        match method:
            case "POST":
                return self._build_post_params()
            case "POST-direct":
                return self._build_post_direct_params()
            case _:
                msg = f"Expected update method 'POST' or 'POST-direct'. Got '{method}'."
                raise ValueError(msg)

    def _build_post_params(self) -> SPARQLOperationParameters:
        return SPARQLOperationParameters(
            method="POST",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data=SPARQLOperationDataMap(
                update=self._update_request,
                version=self._version,
                using_graph_uri=self._using_graph_uri,
                using_named_graph_uri=self._using_named_graph_uri,
            ),
        )

    def _build_post_direct_params(self) -> SPARQLOperationParameters:
        return SPARQLOperationParameters(
            method="POST",
            headers={
                "Content-Type": (
                    "application/sparql-update"
                    f"{'' if self._version is None else f'; version={self._version}'}"
                ),
            },
            content=self._update_request,
            params=SPARQLOperationDataMap(
                using_graph_uri=self._using_graph_uri,
                using_named_graph_uri=self._using_named_graph_uri,
            ),
        )



"""Global fixture definitions for the SPARQLx test suite."""

from collections.abc import Iterator
import time
from typing import Protocol

import pytest
from rdflib import Dataset, Graph

import httpx
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs


class _TriplestoreEndpoints(Protocol):
    sparql_endpoint: str | Graph
    update_endpoint: str | Graph
    graphstore_endpoint: str | Graph

    def __iter__(self) -> Iterator[str]:
        yield from (
            self.sparql_endpoint,
            self.update_endpoint,
            self.graphstore_endpoint,
        )


class OxiGraphEndpoints(_TriplestoreEndpoints):
    """Data Container for Oxigraph SPARQL and Graphstore Endpoints."""

    def __init__(self, host, port):
        self._endpoint_base = f"http://{host}:{port}"

        self.sparql_endpoint = f"{self._endpoint_base}/query"
        self.update_endpoint = f"{self._endpoint_base}/update"
        self.graphstore_endpoint = f"{self._endpoint_base}/store"


def wait_for_service(url: str, timeout: int = 10) -> None:
    for _ in range(10):
        try:
            response = httpx.get(url)
            if response.status_code == 200:
                break
        except httpx.RequestError:
            time.sleep(1)
        else:
            raise RuntimeError(
                f"Requested serivce at {url} "
                f"did not become available after {timeout} seconds."
            )


@pytest.fixture(scope="session")
def oxigraph_service() -> Iterator[OxiGraphEndpoints]:
    """Fixture that starts an Oxigraph Triplestore container and exposes an Endpoint object."""

    with DockerContainer("oxigraph/oxigraph").with_exposed_ports(7878) as container:
        host = container.get_container_host_ip()
        port = container.get_exposed_port(7878)
        oxigraph_endpoints = OxiGraphEndpoints(host=host, port=port)

        wait_for_service(oxigraph_endpoints.sparql_endpoint, timeout=10)
        yield oxigraph_endpoints


@pytest.fixture(scope="function")
def oxigraph_service_with_data(oxigraph_service) -> Iterator[OxiGraphEndpoints]:
    """Dependent Fixture that ingests an RDF graph into a running Oxigraph container."""

    oxigraph_endpoints = oxigraph_service

    with httpx.Client() as client, open("tests/data/test_graphs.trig") as f:
        response = client.put(
            url=oxigraph_endpoints.graphstore_endpoint,
            headers={"Content-Type": "application/trig"},
            content=f.read(),
        )
        response.raise_for_status()

        yield oxigraph_endpoints

        client.delete(
            url=oxigraph_endpoints.graphstore_endpoint,
        )


class FusekiEndpoints(_TriplestoreEndpoints):
    """Data Container for Fuseki SPARQL and Graphstore Endpoints."""

    def __init__(self, host, port):
        self._endpoint_base = f"http://{host}:{port}/ds"

        self.sparql_endpoint = f"{self._endpoint_base}/sparql"
        self.update_endpoint = f"{self._endpoint_base}/update"
        self.graphstore_endpoint = f"{self._endpoint_base}/data"


@pytest.fixture(scope="session")
def fuseki_service() -> Iterator[FusekiEndpoints]:
    """Fixture that starts a Fuseki Triplestore container and exposes an Endpoint object."""

    with (
        DockerContainer("secoresearch/fuseki")
        .with_exposed_ports(3030)
        .with_env("ENABLE_DATA_WRITE", "true")
        .with_env("ENABLE_UPDATE", "true")
    ) as container:
        wait_for_logs(container, "Start Fuseki")

        host = container.get_container_host_ip()
        port = container.get_exposed_port(3030)

        endpoints = FusekiEndpoints(host=host, port=port)
        yield endpoints


@pytest.fixture(scope="function")
def fuseki_service_with_data(fuseki_service) -> Iterator[FusekiEndpoints]:
    """Dependent Fixture that ingests an RDF graph into a running Fuseki container.

    Note: For some reason, Fuseki does not accept DELETE to its GSP endpoint
    for deleting all graphs in the store. However, DROP ALL works just fine.
    """
    auth = httpx.BasicAuth(username="admin", password="pw")

    with httpx.Client(auth=auth) as client, open("tests/data/test_graphs.trig") as f:
        response = client.put(
            url=fuseki_service.graphstore_endpoint,
            content=f.read(),
            headers={"Content-Type": "application/trig"},
        )
        response.raise_for_status()

        yield fuseki_service

        client.post(url=fuseki_service.update_endpoint, data={"update": "drop all"})


class RDFLibGraphEndpoints(_TriplestoreEndpoints):
    """Data Container for Fuseki SPARQL and Graphstore Endpoints."""

    def __init__(self, graph: Graph | None = None):
        _graph = Graph() if graph is None else graph

        self.sparql_endpoint = _graph
        self.update_endpoint = _graph
        self.graphstore_endpoint = _graph


@pytest.fixture(scope="session")
def rdflib_graph_service() -> RDFLibGraphEndpoints:
    return RDFLibGraphEndpoints()


@pytest.fixture(scope="function")
def rdflib_graph_service_with_data() -> RDFLibGraphEndpoints:
    with open("tests/data/test_graphs.trig") as f:
        graph = Dataset().parse(data=f.read(), format="trig")
        return RDFLibGraphEndpoints(graph=graph)


@pytest.fixture(params=["oxigraph_service", "fuseki_service", "rdflib_graph_service"])
def triplestore(request) -> _TriplestoreEndpoints:
    return request.getfixturevalue(request.param)


@pytest.fixture(params=["oxigraph_service_with_data", "fuseki_service_with_data"])
def triplestore_with_data(request) -> _TriplestoreEndpoints:
    return request.getfixturevalue(request.param)



"""Basic tests for SPARQLWrapper.update."""

from typing import NamedTuple

import pytest
from rdflib import Dataset, Graph, URIRef

import httpx
from sparqlx import SPARQLWrapper
from utils import acall, sparql_result_set_equal


class UpdateTestParameter(NamedTuple):
    update_request: str
    update_params: dict
    query: str
    expected: list[dict]


params = [
    UpdateTestParameter(
        update_request="insert data {graph <urn:ng> {<urn:s> <urn:p> <urn:o>}}",
        update_params={},
        query="select ?s ?p ?o where {graph <urn:ng> {?s ?p ?o}}",
        expected=[{"s": URIRef("urn:s"), "p": URIRef("urn:p"), "o": URIRef("urn:o")}],
    ),
    UpdateTestParameter(
        update_request="insert {?s <urn:p> <urn:o>} where {graph <urn:ng1> {?s ?p ?o}}",
        update_params={},
        query="select ?o where {<http://example.org/bob> <urn:p> ?o}",
        expected=[{"o": URIRef("urn:o")}],
    ),
    UpdateTestParameter(
        update_request="insert {?s <urn:p> <urn:o>} where {?s ?p ?o}",
        update_params={"using_graph_uri": "urn:ng1"},
        query="select ?o where {<http://example.org/bob> <urn:p> ?o}",
        expected=[{"o": URIRef("urn:o")}],
    ),
    UpdateTestParameter(
        update_request="""
        insert data {<urn:s> <urn:p> <urn:o>} ;
        insert {graph ?g {<urn:s> <urn:p> <urn:o>}} where {graph ?g {?s ?p ?o}}
        """,
        update_params={},
        query="select ?g ?s ?p ?o where {{<urn:s> ?p ?o} union {graph ?g {<urn:s> ?p ?o}}}",
        expected=[
            {
                "g": URIRef("urn:ng2"),
                "o": URIRef("urn:o"),
                "p": URIRef("urn:p"),
                "s": None,
            },
            {
                "g": URIRef("urn:ng1"),
                "o": URIRef("urn:o"),
                "p": URIRef("urn:p"),
                "s": None,
            },
            {
                "g": None,
                "o": URIRef("urn:o"),
                "p": URIRef("urn:p"),
                "s": None,
            },
        ],
    ),
    UpdateTestParameter(
        update_request="""
        insert data {<urn:s> <urn:p> <urn:o>} ;
        insert {graph ?g {<urn:s> <urn:p> <urn:o>}} where {graph ?g {?s ?p ?o}}
        """,
        update_params={"using_named_graph_uri": "urn:ng1"},
        query="select ?g ?s ?p ?o where {{<urn:s> ?p ?o} union {graph ?g {<urn:s> ?p ?o}}}",
        expected=[
            {
                "g": URIRef("urn:ng1"),
                "o": URIRef("urn:o"),
                "p": URIRef("urn:p"),
                "s": None,
            },
            {
                "g": None,
                "o": URIRef("urn:o"),
                "p": URIRef("urn:p"),
                "s": None,
            },
        ],
    ),
]


@pytest.mark.parametrize("update_method", ["POST", "POST-direct"])
@pytest.mark.parametrize("method", ["update", "aupdate"])
@pytest.mark.parametrize("param", params)
@pytest.mark.asyncio
async def test_sparqlwrapper_update(
    update_method, method, param, oxigraph_service_with_data
):
    sparqlwrapper = SPARQLWrapper(
        sparql_endpoint=oxigraph_service_with_data.sparql_endpoint,
        update_endpoint=oxigraph_service_with_data.update_endpoint,
        update_method=update_method,
    )

    with sparqlwrapper as wrapper:
        result_before_update = wrapper.query(param.query, convert=True)
        assert not result_before_update

        await acall(
            obj=wrapper,
            method=method,
            update_request=param.update_request,
            **param.update_params,
        )

        result_after_update = wrapper.query(param.query, convert=True)
        assert sparql_result_set_equal(result_after_update, param.expected)


@pytest.mark.parametrize("update_method", ["POST", "POST-direct"])
def test_sparqlwrapper_updates(update_method, fuseki_service):
    """Basic test for SPARQLWrapper.updates.

    Note: This test runs against a Fuseki Triplestore to also test auth for updates.
    """
    sparqlwrapper = SPARQLWrapper(
        sparql_endpoint=fuseki_service.sparql_endpoint,
        update_endpoint=fuseki_service.update_endpoint,
        aclient_config={"auth": httpx.BasicAuth(username="admin", password="pm")},
        update_method=update_method,
    )

    sparqlwrapper.updates(
        "insert data {<urn:s> <urn:p> <urn:o>}",
        "insert data {graph <urn:ng1> {<urn:s> <urn:p> <urn:o>}}",
        "insert data {graph <urn:ng2> {<urn:s> <urn:p> <urn:o>}}",
    )

    result = sparqlwrapper.query(
        "select ?g ?s ?p ?o where { {?s ?p ?o} union { graph ?g {?s ?p ?o} }}",
        convert=True,
    )

    expected = [
        {
            "g": None,
            "s": URIRef("urn:s"),
            "p": URIRef("urn:p"),
            "o": URIRef("urn:o"),
        },
        {
            "g": URIRef("urn:ng1"),
            "s": URIRef("urn:s"),
            "p": URIRef("urn:p"),
            "o": URIRef("urn:o"),
        },
        {
            "g": URIRef("urn:ng2"),
            "s": URIRef("urn:s"),
            "p": URIRef("urn:p"),
            "o": URIRef("urn:o"),
        },
    ]

    assert all(_result in expected for _result in result)
    assert all(_expected in result for _expected in expected)


@pytest.mark.asyncio
@pytest.mark.parametrize("method", ["update", "aupdate"])
async def test_sparqlwrapper_update_graph_target(method):
    graph = Graph()
    sparqlwrapper = SPARQLWrapper(sparql_endpoint=graph, update_endpoint=graph)

    assert sparqlwrapper.query("select * where {?s ?p ?o}", convert=True) == []

    await acall(
        sparqlwrapper, method, update_request="insert data {<urn:s> <urn:p> <urn:o>}"
    )

    assert sparqlwrapper.query("select * where {?s ?p ?o}", convert=True) == [
        {"s": URIRef("urn:s"), "p": URIRef("urn:p"), "o": URIRef("urn:o")}
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize("method", ["update", "aupdate"])
async def test_sparqlwrapper_update_dataset_target(method):
    dataset = Dataset()
    sparqlwrapper = SPARQLWrapper(sparql_endpoint=dataset, update_endpoint=dataset)

    assert (
        sparqlwrapper.query("select * where {graph ?g {?s ?p ?o}}", convert=True) == []
    )

    await acall(
        sparqlwrapper,
        method,
        update_request="insert data {graph <urn:g> {<urn:s> <urn:p> <urn:o>}}",
    )

    assert sparqlwrapper.query(
        "select * where {graph ?g {?s ?p ?o}}", convert=True
    ) == [
        {
            "g": URIRef("urn:g"),
            "s": URIRef("urn:s"),
            "p": URIRef("urn:p"),
            "o": URIRef("urn:o"),
        }
    ]



import asyncio
from collections.abc import AsyncIterator
from contextlib import AbstractContextManager, asynccontextmanager
from typing import cast

from rdflib import BNode, Graph
from rdflib.plugins.sparql import prepareQuery
from rdflib.plugins.sparql.sparql import Query
from sparqlx.types import SPARQLQuery, SPARQLQueryTypeLiteral, SPARQLResponseFormat
from sparqlx.utils.converters import _convert_ask, _convert_bindings, _convert_graph


class SPARQLParseException(Exception): ...


def _get_query_type(query: SPARQLQuery) -> SPARQLQueryTypeLiteral:
    try:
        _prepared_query: Query = prepareQuery(query)
    except Exception as exc:
        raise SPARQLParseException(exc) from exc
    else:
        query_type = _prepared_query.algebra.name

    return cast(SPARQLQueryTypeLiteral, query_type)


def _get_response_converter(
    query_type: SPARQLQueryTypeLiteral, response_format: SPARQLResponseFormat | str
):
    if query_type in ["SelectQuery", "AskQuery"] and response_format not in [
        "application/json",
        "application/sparql-results+json",
    ]:
        msg = "JSON response format required for convert=True on SELECT and ASK query results."
        raise ValueError(msg)

    match query_type:
        case "SelectQuery":
            converter = _convert_bindings
        case "AskQuery":
            converter = _convert_ask
        case "DescribeQuery" | "ConstructQuery":
            converter = _convert_graph
        case _:  # pragma: no cover
            raise ValueError(f"Unsupported query type: {query_type}")

    return converter


class Endpoint:
    def __init__(self, endpoint: str | Graph | None):
        self._endpoint = endpoint

    @property
    def url(self) -> str:
        match self._endpoint:
            case str():
                return self._endpoint
            case Graph():
                graph_uri = (
                    identifier.skolemize()
                    if isinstance((identifier := self._endpoint.identifier), BNode)
                    else identifier
                )
                return graph_uri
            case None:
                raise ValueError(
                    "Unable to run SPARQL operations: Endpoint is not defined."
                )
            case _:  # pragma: no cover
                assert False, "This should never happen."

    @property
    def graph(self) -> Graph | None:
        return self._endpoint if isinstance(self._endpoint, Graph) else None


@asynccontextmanager
async def as_async_context[T](sync_cm: AbstractContextManager[T]) -> AsyncIterator[T]:
    """Async context wrapper around a sync context manager.

    The async context manager allows to call a sync context manager
    from an async context statement. This workaround is mentioned in PEP 806,
    see https://peps.python.org/pep-0806/#workaround-an-as-acm-wrapper.
    """
    with sync_cm as result:
        await asyncio.sleep(0)
        yield result



import json
import logging

import httpx


logger = logging.getLogger(__name__)


class StructuredMessage:
    """Simple structured log message class.

    This is taken from the Python logging cookbook:
    https://docs.python.org/3/howto/logging-cookbook.html#implementing-structured-logging
    """

    def __init__(self, message: str, **kwargs):
        self.message = message
        self.kwargs = kwargs

    def __str__(self):
        return "%s >>> %s" % (self.message, json.dumps(self.kwargs, default=str))


def log_request(request: httpx.Request) -> None:
    """Logging event hook for httpx.Requests.

    See httpx Event Hooks:
    https://www.python-httpx.org/advanced/event-hooks/
    """
    info_message = StructuredMessage(
        "Request",
        method=request.method,
        url=request.url,
    )
    logger.info(info_message)

    debug_message = StructuredMessage(
        "Request",
        request=request,
        method=request.method,
        url=request.url,
        content=request.content,
        headers=request.headers,
    )
    logger.debug(debug_message)


def log_response(response: httpx.Response) -> None:
    """Logging event hook for httpx.Responses.

    See httpx Event Hooks:
    https://www.python-httpx.org/advanced/event-hooks/
    """

    info_message = StructuredMessage(
        "Response",
        status_code=response.status_code,
        url=response.url,
    )
    logger.info(info_message)

    debug_message = StructuredMessage(
        "Response",
        status_code=response.status_code,
        reason_phrase=response.reason_phrase,
        http_version=response.http_version,
        url=response.url,
        headers=response.headers,
    )
    logger.debug(debug_message)


async def alog_request(request: httpx.Request) -> None:
    log_request(request=request)


async def alog_response(response: httpx.Response) -> None:
    log_response(response=response)



from collections.abc import AsyncIterator, Iterator
from contextlib import asynccontextmanager, contextmanager
from typing import Literal
import warnings

import httpx
from sparqlx.utils.logging_hooks import (
    alog_request,
    alog_response,
    log_request,
    log_response,
)


class ClientManager:
    def __init__(
        self,
        client: httpx.Client | None = None,
        client_config: dict | None = None,
        aclient: httpx.AsyncClient | None = None,
        aclient_config: dict | None = None,
    ) -> None:
        self._client = client
        self._client_config = client_config
        self._aclient = aclient
        self._aclient_config = aclient_config

    @property
    def client(self) -> httpx.Client:
        return self._get_client(client=self._client, client_config=self._client_config)

    @property
    def aclient(self) -> httpx.AsyncClient:
        return self._get_aclient(
            aclient=self._aclient, aclient_config=self._aclient_config
        )

    @contextmanager
    def context(self) -> Iterator[httpx.Client]:
        client = self.client

        try:
            yield client
        finally:
            if self._client is None:
                client.close()
            else:
                self._open_client_warning(client)

    @asynccontextmanager
    async def acontext(self) -> AsyncIterator[httpx.AsyncClient]:
        aclient = self.aclient

        try:
            yield aclient
        finally:
            if self._aclient is None:
                await aclient.aclose()
            else:
                self._open_client_warning(aclient)

    @staticmethod
    def _open_client_warning(client: httpx.Client | httpx.AsyncClient) -> None:
        msg = (
            f"httpx Client instance '{client}' is not managed. "
            "Client.close/AsyncClient.aclose should be called at some point."
        )
        warnings.warn(msg, stacklevel=2)

    @staticmethod
    def _add_event_hook(
        client: httpx.Client | httpx.AsyncClient,
        hook_type: Literal["request", "response"],
        hook: httpx._client.EventHook,
    ) -> None:
        if hook not in (hooks := client.event_hooks[hook_type]):
            hooks.append(hook)

    def _get_client(
        self, client: httpx.Client | None, client_config: dict | None
    ) -> httpx.Client:
        client = httpx.Client(**(client_config or {})) if client is None else client

        self._add_event_hook(client, "request", log_request)
        self._add_event_hook(client, "response", log_response)

        return client

    def _get_aclient(
        self, aclient: httpx.AsyncClient | None, aclient_config: dict | None
    ) -> httpx.AsyncClient:
        aclient = (
            httpx.AsyncClient(**(aclient_config or {})) if aclient is None else aclient
        )

        self._add_event_hook(aclient, "request", alog_request)
        self._add_event_hook(aclient, "response", alog_response)

        return aclient



"""Basic sad path tests for SPARQLWrapper."""

from typing import NamedTuple

import pytest

from data.queries import ask_query_false, ask_query_true, select_query_xy_values
from sparqlx import SPARQLParseException, SPARQLWrapper
from sparqlx.utils.operation_parameters import sparql_result_response_format_map


@pytest.mark.parametrize(
    "query", [select_query_xy_values, ask_query_true, ask_query_false]
)
@pytest.mark.parametrize(
    "response_format",
    filter(lambda k: k != "json", sparql_result_response_format_map.keys()),
)
def test_sparqlwrapper_result_binding_conversion_non_json_fail(
    query, response_format, triplestore
):
    msg = "JSON response format required for convert=True on SELECT and ASK query results."
    with pytest.raises(ValueError, match=msg):
        SPARQLWrapper(sparql_endpoint=triplestore.sparql_endpoint).query(
            query, convert=True, response_format=response_format
        )


class SPARQLParseExceptionTestParameter(NamedTuple):
    invalid_input: str
    exception_match_text: str


params = [
    SPARQLParseExceptionTestParameter(
        invalid_input="NOT A SPARQL QUERY",
        exception_match_text="Expected {SelectQuery | ConstructQuery | DescribeQuery | AskQuery}, found 'NOT'",
    ),
    SPARQLParseExceptionTestParameter(
        invalid_input="select * where {?s ?p ?o ",
        exception_match_text="Expected SelectQuery, found end of text",
    ),
    SPARQLParseExceptionTestParameter(
        invalid_input="ask ?s where {?s ?p ?o}",
        exception_match_text="Expected AskQuery, found '?'",
    ),
]


@pytest.mark.parametrize("param", params)
def test_sparqlwrapper_parse_exception(param, triplestore):
    """Simple test case for invalid SPARQL query inputs.

    Check that an invalid SPARQL query input raises SPARQLParseException and
    that the pyparsing.ParseException message is propagated to SPARQLParseException.
    """
    with pytest.raises(SPARQLParseException, match=param.exception_match_text):
        SPARQLWrapper(sparql_endpoint=triplestore.sparql_endpoint).query(
            param.invalid_input
        )


def test_sparqlwrapper_invalid_config():
    """Check that SPARQLWrapper init without any endpoint definition fails."""

    msg = (
        "Invalid SPARQLWrapper configuration: "
        "at least one of 'sparql_endpoint' or 'update_endpoint' must be set."
    )

    with pytest.raises(ValueError, match=msg):
        SPARQLWrapper()


def test_sparqlwrapper_no_endpoints():
    fail_msg = (
        "Invalid SPARQLWrapper configuration: "
        "at least one of 'sparql_endpoint' or 'update_endpoint' must be set."
    )

    with pytest.raises(ValueError, match=fail_msg):
        SPARQLWrapper()


def test_sparqlwrapper_undefined_endpoints():
    fail_msg = "Unable to run SPARQL operations: Endpoint is not defined."

    sparqlwrapper_query = SPARQLWrapper(update_endpoint="...")
    sparqlwrapper_update = SPARQLWrapper(sparql_endpoint="...")

    with pytest.raises(ValueError, match=fail_msg):
        sparqlwrapper_query.query("select * where {?s ?p ?o}")

    with pytest.raises(ValueError, match=fail_msg):
        sparqlwrapper_update.update("insert data {<urn:s> <urn:p> <urn:o>}")



"""Pytest entry point for sparqlx.SPARQLWrapper URI parameter tests."""

import itertools
from typing import NamedTuple

import pytest
from rdflib import URIRef

from conftest import FusekiEndpoints
from sparqlx import SPARQLWrapper
from utils import sparql_result_set_equal


class URITestParameter(NamedTuple):
    query: str
    expected: list[dict]

    default_graph_uri: str | list[str] | None = None
    named_graph_uri: str | list[str] | None = None


count_triples_query = "select (count(*) as ?cnt) where {?s ?p ?o}"
select_graphs_query = "select distinct ?g where {graph ?g {?s ?p ?o}}"

default_graph_uri_params = [
    URITestParameter(
        query=count_triples_query, expected=[{"cnt": 33}], default_graph_uri=None
    ),
    URITestParameter(
        query=count_triples_query, expected=[{"cnt": 2}], default_graph_uri="urn:ng1"
    ),
    URITestParameter(
        query=count_triples_query,
        expected=[{"cnt": 4}],
        default_graph_uri=["urn:ng1", "urn:ng2"],
    ),
]

named_graph_uri_params = [
    URITestParameter(
        query=select_graphs_query,
        expected=[
            {"g": URIRef("urn:ng2")},
            {"g": URIRef("urn:ng1")},
        ],
    ),
    URITestParameter(
        query=select_graphs_query,
        named_graph_uri="urn:ng1",
        expected=[
            {"g": URIRef("urn:ng1")},
        ],
    ),
    URITestParameter(
        query=select_graphs_query,
        named_graph_uri=["urn:ng1", "urn:ng2"],
        expected=[
            {"g": URIRef("urn:ng1")},
            {"g": URIRef("urn:ng2")},
        ],
    ),
]

mixed_uri_parameters = [
    URITestParameter(
        query=count_triples_query,
        expected=[{"cnt": 4}],
        default_graph_uri=["urn:ng1", "urn:ng2"],
        named_graph_uri="urn:ng1",
    ),
    URITestParameter(
        query=select_graphs_query,
        expected=[{"g": URIRef("urn:ng1")}],
        default_graph_uri=["urn:ng1", "urn:ng2"],
        named_graph_uri="urn:ng1",
    ),
]


@pytest.mark.parametrize(
    "params",
    itertools.chain(
        default_graph_uri_params, named_graph_uri_params, mixed_uri_parameters
    ),
)
def test_sparqlwrapper_default_graph_uri(params, triplestore_with_data):
    if (
        isinstance(triplestore_with_data, FusekiEndpoints)
        and params == default_graph_uri_params[0]
    ):
        reason = """The Fuseki store actually cannot implicitely access the default graph.

        The secoresearch/fuseki image unfortunately sets the
        tdb:unionDefaultGraph setting to true, which makes the union of
        named graphs the default graph and excludes the actual default graph data.

        See https://jena.apache.org/documentation/tdb/datasets.html (Dataset Query).
        """
        pytest.skip(reason=reason)

    sparqlwrapper = SPARQLWrapper(sparql_endpoint=triplestore_with_data.sparql_endpoint)
    result = sparqlwrapper.query(
        query=params.query,
        convert=True,
        default_graph_uri=params.default_graph_uri,
        named_graph_uri=params.named_graph_uri,
    )

    assert sparql_result_set_equal(result, params.expected)



"""SPARQLWrapper: An httpx-based SPARQL 1.2 Protocol client."""

import asyncio
from collections.abc import AsyncIterator, Callable, Iterator
from contextlib import AbstractAsyncContextManager, AbstractContextManager
import functools
from typing import Literal as TLiteral, Self, overload
import warnings

import httpx
from rdflib import Dataset, Graph
from sparqlx.types import (
    AskQuery,
    ConstructQuery,
    DescribeQuery,
    RDFParseSource,
    RequestDataValue,
    SPARQLQuery,
    SPARQLQueryTypeLiteral,
    SPARQLResponseFormat,
    SPARQLResultBinding,
    SelectQuery,
)
from sparqlx.utils.client_manager import ClientManager
from sparqlx.utils.operation_parameters import (
    QueryOperationParametersConstructor,
    SPARQLOperationParameters,
    UpdateOperationParametersConstructor,
)
from sparqlx.utils.transports import (
    AsyncRDFLibQueryTransport,
    AsyncRDFLibUpdateTransport,
    RDFLibQueryTransport,
    RDFLibUpdateTransport,
)
from sparqlx.utils.utils import (
    Endpoint,
    _get_query_type,
    _get_response_converter,
    as_async_context,
)


class SPARQLWrapper(AbstractContextManager, AbstractAsyncContextManager):
    """SPARQLWrapper: An httpx-based SPARQL 1.2 Protocol client.

    The class provides functionality for running SPARQL Query and Update Operations
    according to the SPARQL 1.2 protocol and supports both sync and async interfaces.
    """

    def __init__(
        self,
        sparql_endpoint: str | Graph | None = None,
        update_endpoint: str | Graph | None = None,
        client: httpx.Client | None = None,
        client_config: dict | None = None,
        aclient: httpx.AsyncClient | None = None,
        aclient_config: dict | None = None,
        query_method: TLiteral["GET", "POST", "POST-direct"] = "POST",
        update_method: TLiteral["POST", "POST-direct"] = "POST",
    ) -> None:
        if sparql_endpoint is None and update_endpoint is None:
            raise ValueError(
                "Invalid SPARQLWrapper configuration: "
                "at least one of 'sparql_endpoint' or 'update_endpoint' must be set."
            )
        self._sparql_endpoint = Endpoint(sparql_endpoint)
        self._update_endpoint = Endpoint(update_endpoint)

        self._query_method = query_method
        self._update_method = update_method

        self._client_config: dict = client_config or {}
        self._aclient_config: dict = aclient_config or {}

        self._client_manager = ClientManager(
            client=client,
            client_config=client_config,
            aclient=aclient,
            aclient_config=aclient_config,
        )

    @classmethod
    def from_rdf_source(
        cls, rdf_source: RDFParseSource | Graph, **parse_kwargs
    ) -> Self:
        """Alternative constructor for instantiating a SPARQLWrapper from an RDF source.

        The constructor instantiates a sparqlx.SPARQLWrapper with an rdflib.Graph target;
        the target for both sparql_endpoint and update_endpoint will be either an rdflib.Dataset
        obtained by parsing an RDFParseSource or an rdflib.Graph object passed to the constructor.

        kwargs are forwarded to the rdflib.Graph.parse methods.

        Note that the constructor does not allow to exert control over the constructed SPARQLWrapper;
        the SPARQLWrapper config concerns HTTP targets (clients, client configs and HTTP methods)
        and are effectively meaningless for rdflib.Graph targets.
        """

        graph: Graph = (
            rdf_source
            if isinstance(rdf_source, Graph)
            else Dataset().parse(source=rdf_source, **parse_kwargs)
        )

        return cls(sparql_endpoint=graph, update_endpoint=graph)

    def __enter__(self) -> Self:
        self._client_manager._client = self._client_manager.client
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._client_manager.client.close()

    async def __aenter__(self) -> Self:
        self._client_manager._aclient = self._client_manager.aclient
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self._client_manager.aclient.aclose()

    @overload
    def query(
        self,
        query: SelectQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> list[SPARQLResultBinding]: ...

    @overload
    def query(
        self,
        query: AskQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> bool: ...

    @overload
    def query(
        self,
        query: ConstructQuery | DescribeQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> Graph: ...

    @overload
    def query(
        self,
        query: SPARQLQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> list[SPARQLResultBinding] | Graph | bool: ...

    @overload
    def query(
        self,
        query: SPARQLQuery,
        convert: TLiteral[False] = False,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> httpx.Response: ...

    def query(
        self,
        query: SPARQLQuery,
        convert: bool = False,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> httpx.Response | list[SPARQLResultBinding] | Graph | bool:
        query_type: SPARQLQueryTypeLiteral = _get_query_type(query=query)

        params: SPARQLOperationParameters = QueryOperationParametersConstructor(
            query=query,
            query_type=query_type,
            response_format=response_format,
            version=version,
            default_graph_uri=default_graph_uri,
            named_graph_uri=named_graph_uri,
        ).get_params(method=self._query_method)  # pyright: ignore

        response_handler = (
            _get_response_converter(
                query_type=query_type, response_format=params.headers["Accept"]
            )
            if convert
            else lambda response: response
        )

        client_context: AbstractContextManager[httpx.Client] = (
            self._client_manager.context()
            if (graph := self._sparql_endpoint.graph) is None
            else httpx.Client(
                **self._client_config,
                transport=RDFLibQueryTransport(
                    query=query,
                    graph=graph,
                    version=version,
                    default_graph_uri=default_graph_uri,
                    named_graph_uri=named_graph_uri,
                ),
            )
        )

        with client_context as client:
            response = client.request(
                method=params.method,
                url=self._sparql_endpoint.url,
                content=params.content,
                data=params.data,
                headers=params.headers,
                params=params.params,
            )

            if convert or raise_for_status:
                response.raise_for_status()

        return response_handler(response=response)

    @overload
    async def aquery(
        self,
        query: SelectQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> list[SPARQLResultBinding]: ...

    @overload
    async def aquery(
        self,
        query: AskQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> bool: ...

    @overload
    async def aquery(
        self,
        query: ConstructQuery | DescribeQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> Graph: ...

    @overload
    async def aquery(
        self,
        query: SPARQLQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> list[SPARQLResultBinding] | Graph | bool: ...

    @overload
    async def aquery(
        self,
        query: SPARQLQuery,
        convert: TLiteral[False] = False,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> httpx.Response: ...

    async def aquery(
        self,
        query: SPARQLQuery,
        convert: bool = False,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> httpx.Response | list[SPARQLResultBinding] | Graph | bool:
        query_type: SPARQLQueryTypeLiteral = _get_query_type(query=query)

        params: SPARQLOperationParameters = QueryOperationParametersConstructor(
            query=query,
            query_type=query_type,
            response_format=response_format,
            version=version,
            default_graph_uri=default_graph_uri,
            named_graph_uri=named_graph_uri,
        ).get_params(method=self._query_method)  # pyright: ignore

        response_handler = (
            _get_response_converter(
                query_type=query_type, response_format=params.headers["Accept"]
            )
            if convert
            else lambda response: response
        )

        aclient_context: AbstractAsyncContextManager[httpx.AsyncClient] = (
            self._client_manager.acontext()
            if (graph := self._sparql_endpoint.graph) is None
            else httpx.AsyncClient(
                **self._client_config,
                transport=AsyncRDFLibQueryTransport(
                    query=query,
                    graph=graph,
                    version=version,
                    default_graph_uri=default_graph_uri,
                    named_graph_uri=named_graph_uri,
                ),
            )
        )

        async with aclient_context as aclient:
            response = await aclient.request(
                method=params.method,
                url=self._sparql_endpoint.url,
                content=params.content,
                data=params.data,
                headers=params.headers,
                params=params.params,
            )

            if convert or raise_for_status:
                response.raise_for_status()

        return response_handler(response=response)

    def query_stream[T](
        self,
        query: SPARQLQuery,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        streaming_method: Callable[
            [httpx.Response], Iterator[T]
        ] = httpx.Response.iter_bytes,
        chunk_size: int | None = None,
    ) -> Iterator[T]:
        if self._sparql_endpoint.graph is not None:
            msg = "Response streaming is currently not supported for rdflib.Graph targets."
            raise NotImplementedError(msg)

        query_type: SPARQLQueryTypeLiteral = _get_query_type(query=query)

        params: SPARQLOperationParameters = QueryOperationParametersConstructor(
            query=query,
            query_type=query_type,
            response_format=response_format,
            version=version,
            default_graph_uri=default_graph_uri,
            named_graph_uri=named_graph_uri,
        ).get_params(method=self._query_method)  # pyright: ignore

        _streaming_method = (
            streaming_method
            if chunk_size is None
            else functools.partial(streaming_method, chunk_size=chunk_size)  # type: ignore
        )

        with self._client_manager.context() as client:
            with client.stream(
                method=params.method,
                url=self._sparql_endpoint.url,
                content=params.content,
                data=params.data,
                headers=params.headers,
                params=params.params,
            ) as response:
                response.raise_for_status()

                for chunk in _streaming_method(response):
                    yield chunk

    async def aquery_stream[T](
        self,
        query: SPARQLQuery,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        streaming_method: Callable[
            [httpx.Response], AsyncIterator[T]
        ] = httpx.Response.aiter_bytes,
        chunk_size: int | None = None,
    ) -> AsyncIterator[T]:
        if self._sparql_endpoint.graph is not None:
            msg = "Response streaming is currently not supported for rdflib.Graph targets."
            raise NotImplementedError(msg)

        query_type: SPARQLQueryTypeLiteral = _get_query_type(query=query)

        params: SPARQLOperationParameters = QueryOperationParametersConstructor(
            query=query,
            query_type=query_type,
            response_format=response_format,
            version=version,
            default_graph_uri=default_graph_uri,
            named_graph_uri=named_graph_uri,
        ).get_params(method=self._query_method)  # pyright: ignore

        _streaming_method = (
            streaming_method
            if chunk_size is None
            else functools.partial(streaming_method, chunk_size=chunk_size)  # type: ignore
        )

        async with self._client_manager.acontext() as aclient:
            async with aclient.stream(
                method=params.method,
                url=self._sparql_endpoint.url,
                content=params.content,
                data=params.data,
                headers=params.headers,
                params=params.params,
            ) as response:
                response.raise_for_status()

                async for chunk in _streaming_method(response):
                    yield chunk

    @overload
    def queries(
        self,
        *queries: SPARQLQuery,
        convert: TLiteral[True],
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> Iterator[list[SPARQLResultBinding] | Graph | bool]: ...

    @overload
    def queries(
        self,
        *queries: SPARQLQuery,
        convert: TLiteral[False] = False,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> Iterator[httpx.Response]: ...

    def queries(
        self,
        *queries: SPARQLQuery,
        convert: bool = False,
        response_format: SPARQLResponseFormat | str | None = None,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> Iterator[httpx.Response | list[SPARQLResultBinding] | Graph | bool]:
        query_component = SPARQLWrapper(
            sparql_endpoint=self._sparql_endpoint._endpoint,
            aclient=self._client_manager.aclient,
            query_method=self._query_method,  # pyright: ignore
        )

        async def _runner() -> Iterator[httpx.Response]:
            acatch_warnings: AbstractAsyncContextManager[None] = as_async_context(
                warnings.catch_warnings()
            )

            async with (
                acatch_warnings,
                query_component,
                asyncio.TaskGroup() as tg,
            ):
                warnings.filterwarnings(
                    action="ignore",
                    category=UserWarning,
                    module="sparqlx.utils.client_manager",
                )

                tasks = [
                    tg.create_task(
                        query_component.aquery(
                            query=query,
                            convert=convert,
                            response_format=response_format,
                            version=version,
                            default_graph_uri=default_graph_uri,
                            named_graph_uri=named_graph_uri,
                            raise_for_status=raise_for_status,
                        )
                    )
                    for query in queries
                ]

            return map(asyncio.Task.result, tasks)

        results = asyncio.run(_runner())
        return results

    def update(
        self,
        update_request: str,
        version: str | None = None,
        using_graph_uri: RequestDataValue = None,
        using_named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        params: SPARQLOperationParameters = UpdateOperationParametersConstructor(
            update_request=update_request,
            version=version,
            using_graph_uri=using_graph_uri,
            using_named_graph_uri=using_named_graph_uri,
        ).get_params(method=self._update_method)  # pyright: ignore

        client_context: AbstractContextManager[httpx.Client] = (
            self._client_manager.context()
            if (graph := self._update_endpoint.graph) is None
            else httpx.Client(
                **self._client_config,
                transport=RDFLibUpdateTransport(
                    update_request=update_request,
                    graph=graph,
                    version=version,
                    using_graph_uri=using_graph_uri,
                    using_named_graph_uri=using_named_graph_uri,
                ),
            )
        )

        with client_context as client:
            response = client.request(
                method=params.method,
                url=self._update_endpoint.url,
                content=params.content,
                data=params.data,
                headers=params.headers,
                params=params.params,
            )

            if raise_for_status:
                response.raise_for_status()

            return response

    async def aupdate(
        self,
        update_request: str,
        version: str | None = None,
        using_graph_uri: RequestDataValue = None,
        using_named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        params: SPARQLOperationParameters = UpdateOperationParametersConstructor(
            update_request=update_request,
            version=version,
            using_graph_uri=using_graph_uri,
            using_named_graph_uri=using_named_graph_uri,
        ).get_params(method=self._update_method)  # pyright: ignore

        aclient_context: AbstractAsyncContextManager[httpx.AsyncClient] = (
            self._client_manager.acontext()
            if (graph := self._update_endpoint.graph) is None
            else httpx.AsyncClient(
                **self._aclient_config,
                transport=AsyncRDFLibUpdateTransport(
                    update_request=update_request,
                    graph=graph,
                    version=version,
                    using_graph_uri=using_graph_uri,
                    using_named_graph_uri=using_named_graph_uri,
                ),
            )
        )

        async with aclient_context as aclient:
            response = await aclient.request(
                method=params.method,
                url=self._update_endpoint.url,
                content=params.content,
                data=params.data,
                headers=params.headers,
                params=params.params,
            )

            if raise_for_status:
                response.raise_for_status()

            return response

    def updates(
        self,
        *update_requests,
        version: str | None = None,
        using_graph_uri: RequestDataValue = None,
        using_named_graph_uri: RequestDataValue = None,
        raise_for_status: bool = True,
    ) -> Iterator[httpx.Response]:
        update_component = SPARQLWrapper(
            update_endpoint=self._update_endpoint._endpoint,
            aclient=self._client_manager.aclient,
            update_method=self._update_method,  # pyright: ignore
        )

        async def _runner() -> Iterator[httpx.Response]:
            async with update_component, asyncio.TaskGroup() as tg:
                tasks = [
                    tg.create_task(
                        update_component.aupdate(
                            update_request=update_request,
                            version=version,
                            using_graph_uri=using_graph_uri,
                            using_named_graph_uri=using_named_graph_uri,
                            raise_for_status=raise_for_status,
                        )
                    )
                    for update_request in update_requests
                ]

            return map(asyncio.Task.result, tasks)

        results = asyncio.run(_runner())
        return results



"""SPARQL queries and update requests for SPARQLx testing."""

select_query_xy_values = """
    select ?x ?y
    where {
      values (?x ?y) {
        (1 2)
        (3 4)
      }
    }
"""

select_query_types = """
prefix xsd: <http://www.w3.org/2001/XMLSchema#>

select *
where {
  values (?x) {
     (2)
     (2.2)
     (UNDEF)
     (<https://test.uri>)
     ('2024-01-01'^^xsd:date)
     ('2024'^^xsd:gYear)
     ('2024-01'^^xsd:gYearMonth)
    }
}
"""

select_query_bnode = """
select *
where {
    bind (BNODE() as ?x)
}
"""

ask_query_true = """
    ask where {
      values (?x ?y) {
        (1 2)
        (3 4)
      }
    }
"""

ask_query_false = "ask where {filter(false)}"


construct_query_x_values = """
    construct {<urn:s> <urn:p> ?x}
    where {
    values ?x {
        1 2 3
      }
    }
"""

describe_query = "describe ?s where {?s ?p ?o}"



"""SPARQLWrapper type definitions."""

import datetime
import decimal
from pathlib import PurePath
from typing import IO, TextIO
from typing import Literal as PyLiteral
from xml.dom.minidom import Document

from rdflib import BNode, Literal, URIRef
from rdflib.compat import long_type
from rdflib.parser import InputSource
from rdflib.xsd_datetime import Duration


__all__ = (
    "RequestDataValue",
    "LiteralToPython",
    "SPARQLResultBindingValue",
    "SPARQLResultBinding",
    "SPARQLResultResponseFormat",
    "RDFResponseFormat",
    "SPARQLResponseFormat",
    "SPARQLQueryType",
    "SelectQuery",
    "AskQuery",
    "ConstructQuery",
    "DescribeQuery",
    "SPARQLQuery",
    "SPARQLQueryTypeLiteral",
    "RDFParseSource",
)

type RDFParseSource = IO[bytes] | TextIO | InputSource | str | bytes | PurePath
"""Source parameter type for rdflib.Graph.parse.

This is the exact union type as defined in RDFLib.
"""

type RequestDataValue = str | list[str] | tuple[str, ...] | None
"""Value type for httpx.RequestValue type.

See the httpx URL encoding function:
https://github.com/encode/httpx/blob/4b23574cf83307ce27d3b14b4a425dc58c57d28d/httpx/_content.py#L136
"""

type LiteralToPython = (
    Literal
    | None
    | datetime.date
    | datetime.datetime
    | datetime.time
    | datetime.timedelta
    | Duration
    | bytes
    | bool
    | int
    | float
    | decimal.Decimal
    | long_type
    | Document
)
"""Return type for rdflib.Literal.toPython.

This union type represents all possible return value types of Literal.toPython.
Return type provenance:

    - Literal: rdflib.Literal.toPython
    - None: rdflib.term._castLexicalToPython

    - datetime.datetime: rdflib.xsd_datetime.parse_datetime
    - datetime.time: rdflib.xsd_datetime.parse_time
    - datetime.timedelta, Duration: parse_xsd_duration
    - bytes: rdflib.term._unhexlify, base64.b64decode
    - bool: rdflib.term._parseBoolean
    - int, float, decimal.Decimal, long_type: rdflib.term.XSDToPython
    - Document: rdflib.term._parseXML
"""


type SPARQLResultBindingValue = URIRef | BNode | LiteralToPython
"Return type for SPARQLWrapper result mapping values."

type SPARQLResultBinding = dict[str, SPARQLResultBindingValue]

type SPARQLResultResponseFormat = PyLiteral["json", "xml", "csv", "tsv"]
type RDFResponseFormat = PyLiteral["turtle", "xml", "ntriples", "json-ld"]
type SPARQLResponseFormat = SPARQLResultResponseFormat | RDFResponseFormat


class SPARQLQueryType(str): ...


class SelectQuery(SPARQLQueryType): ...


class AskQuery(SPARQLQueryType): ...


class ConstructQuery(SPARQLQueryType): ...


class DescribeQuery(SPARQLQueryType): ...


type SPARQLQuery = str | SelectQuery | AskQuery | ConstructQuery | DescribeQuery


type SPARQLQueryTypeLiteral = PyLiteral[
    "SelectQuery", "AskQuery", "ConstructQuery", "DescribeQuery"
]



import asyncio
from contextlib import suppress
import random

import httpx
import pytest
from sparqlx.utils.client_manager import ClientManager

from utils import as_async_cm


def test_client_manager_automanaged_client():
    client_manager = ClientManager()

    assert client_manager.client is not client_manager.client
    assert client_manager._client is None

    with client_manager.context() as client:
        assert not client.is_closed
    assert client.is_closed

    with client_manager.context() as client1:
        assert not client1.is_closed
        with client_manager.context() as client2:
            assert client1 is not client2

            assert not client1.is_closed
            assert not client2.is_closed

        assert not client1.is_closed
        assert client2.is_closed
    assert client1.is_closed


def test_client_shared_client():
    shared_client = httpx.Client()
    client_manager = ClientManager(client=shared_client)

    assert client_manager.client is shared_client
    assert client_manager.client is client_manager.client
    assert client_manager._client is not None

    with client_manager.context() as client:
        assert client is shared_client
        assert client is client_manager.client

        assert not client.is_closed
        assert not shared_client.is_closed

    assert not client.is_closed
    assert not shared_client.is_closed

    with client_manager.context() as client1:
        assert client1 is shared_client
        assert client1 is client_manager.client

        assert not client1.is_closed

        with client_manager.context() as client2:
            assert client1 is shared_client
            assert client1 is client_manager.client
            assert client1 is client2

            assert not client1.is_closed
            assert not client2.is_closed

        assert not client1.is_closed
        assert not client2.is_closed

    assert not client1.is_closed
    assert not client2.is_closed


@pytest.mark.asyncio
async def test_client_manager_automanaged_aclient():
    client_manager = ClientManager()

    assert client_manager.aclient is not client_manager.aclient
    assert client_manager._aclient is None

    async with client_manager.acontext() as aclient:
        assert not aclient.is_closed
    assert aclient.is_closed

    async with client_manager.acontext() as aclient1:
        assert not aclient1.is_closed
        async with client_manager.acontext() as aclient2:
            assert aclient1 is not aclient2

            assert not aclient1.is_closed
            assert not aclient2.is_closed

        assert not aclient1.is_closed
        assert aclient2.is_closed
    assert aclient1.is_closed


@pytest.mark.asyncio
async def test_client_manager_shared_aclient():
    shared_aclient = httpx.AsyncClient()
    client_manager = ClientManager(aclient=shared_aclient)

    assert client_manager.aclient is shared_aclient
    assert client_manager.aclient is client_manager.aclient
    assert client_manager._aclient is not None

    async with client_manager.acontext() as aclient:
        assert aclient is shared_aclient
        assert aclient is client_manager.aclient

        assert not aclient.is_closed
        assert not shared_aclient.is_closed

    assert not aclient.is_closed
    assert not shared_aclient.is_closed

    async with client_manager.acontext() as aclient1:
        assert aclient1 is shared_aclient
        assert aclient1 is client_manager.aclient

        assert not aclient1.is_closed

        async with client_manager.acontext() as aclient2:
            assert aclient1 is shared_aclient
            assert aclient1 is client_manager.aclient
            assert aclient1 is aclient2

            assert not aclient1.is_closed
            assert not aclient2.is_closed

        assert not aclient1.is_closed
        assert not aclient2.is_closed

    assert not aclient1.is_closed
    assert not aclient2.is_closed


@pytest.mark.asyncio
async def test_client_manager_automanaged_aclient_multi_coros():
    client_manager = ClientManager()

    async def coro(tag):
        async with client_manager.acontext() as aclient:
            assert not aclient.is_closed

            await asyncio.sleep(random.uniform(0, 0.1))

            assert not aclient.is_closed

        assert aclient.is_closed
        return tag

    _range = list(range(5))
    result = await asyncio.gather(*[coro(i) for i in _range])

    assert result == _range


@pytest.mark.asyncio
async def test_client_manager_shared_aclient_multi_coros():
    shared_aclient = httpx.AsyncClient()
    client_manager = ClientManager(aclient=shared_aclient)

    async def coro(tag):
        async with client_manager.acontext() as aclient:
            assert aclient is shared_aclient
            assert not aclient.is_closed

            await asyncio.sleep(random.uniform(0, 0.1))

            assert not aclient.is_closed
        assert not aclient.is_closed
        return tag

    _range = list(range(5))
    result = await asyncio.gather(*[coro(i) for i in _range])

    assert result == _range
    assert not shared_aclient.is_closed

    await shared_aclient.aclose()
    assert shared_aclient.is_closed


def test_client_manager_cleanup_automanaged_client():
    """Check that ClientManager.context closes an automanaged client if an exceptional state occurs."""
    client_manager = ClientManager()

    with suppress(RuntimeError), client_manager.context() as client:
        raise RuntimeError

    assert client.is_closed


def test_client_manager_cleanup_shared_client():
    """Check that ClientManager.context does NOT close a shared client if an exceptional state occurs."""
    shared_client = httpx.Client()
    client_manager = ClientManager(client=shared_client)

    with suppress(RuntimeError), client_manager.context() as client:
        raise RuntimeError

    assert not client.is_closed


@pytest.mark.asyncio
async def test_client_manager_cleanup_automanaged_aclient():
    """Check that ClientManager.acontext closes an automanaged aclient if an exceptional state occurs."""
    client_manager = ClientManager()

    async with (
        as_async_cm(suppress(RuntimeError)),
        client_manager.acontext() as aclient,
    ):
        raise RuntimeError

    assert aclient.is_closed


@pytest.mark.asyncio
async def test_client_manager_cleanup_shared_aclient():
    """Check that ClientManager.acontext does NOT close a shared aclient if an exceptional state occurs."""
    shared_aclient = httpx.AsyncClient()
    client_manager = ClientManager(aclient=shared_aclient)

    async with (
        as_async_cm(suppress(RuntimeError)),
        client_manager.acontext() as aclient,
    ):
        assert aclient is shared_aclient
        assert not aclient.is_closed

        raise RuntimeError

    assert not aclient.is_closed



"""Basic unit tests for StructuredMessage class."""

import json
from typing import Annotated, Any, NamedTuple

import pytest
from sparqlx.utils.logging_hooks import StructuredMessage


class StructuredMessageParams(NamedTuple):
    message: str
    kwargs: dict[str, Any]
    expected: dict[str, Any] | None = None


obj: Annotated[
    object, """Shared instance serving as a non-JSON-serializable object."""
] = object()


params = [
    StructuredMessageParams(
        message="Logging Message",
        kwargs={"x": 1, "y": "2"},
    ),
    StructuredMessageParams(
        message="Logging Message",
        kwargs={"x": 1, "y": None, "obj": obj},
        expected={"x": 1, "y": None, "obj": str(obj)},
    ),
    StructuredMessageParams(
        message="Logging Message",
        kwargs={"x": 1, "y": "y >>> z"},
    ),
]


@pytest.mark.parametrize("param", params)
def test_structured_message(param):
    struct_msg = StructuredMessage(message=param.message, **param.kwargs)
    message, structured = str(struct_msg).split(" >>> ", maxsplit=1)

    assert message == param.message
    assert (param.expected or param.kwargs) == json.loads(structured)



"""Pytest entry point for SPARQLWrapper httpx client sharing."""

import httpx
from sparqlx import SPARQLWrapper


def test_client_identity():
    """Check the client attribute of a SPARQLOperationWrapper for supplied/managed client."""
    sparqlwrapper = SPARQLWrapper(sparql_endpoint="https://some.endpoint")

    client = httpx.Client()
    aclient = httpx.AsyncClient()
    sparqlwrapper_with_client = SPARQLWrapper(
        sparql_endpoint="https://some.endpoint", client=client, aclient=aclient
    )

    assert sparqlwrapper._client_manager._client is None
    assert sparqlwrapper._client_manager._aclient is None

    assert isinstance(sparqlwrapper._client_manager.client, httpx.Client)
    assert isinstance(sparqlwrapper._client_manager.aclient, httpx.AsyncClient)

    assert sparqlwrapper_with_client._client_manager.client is client
    assert sparqlwrapper_with_client._client_manager.aclient is aclient

    assert sparqlwrapper_with_client._client_manager._client is client
    assert sparqlwrapper_with_client._client_manager._aclient is aclient


def test_shared_client():
    """Check client identity and closed-status for a shared client."""
    client = httpx.Client()

    sparqlwrapper_1 = SPARQLWrapper(
        sparql_endpoint="https://some.endpoint", client=client
    )
    sparqlwrapper_2 = SPARQLWrapper(
        sparql_endpoint="https://some.endpoint", client=client
    )

    assert all(
        _client is client
        for _client in [
            sparqlwrapper_1._client_manager.client,
            sparqlwrapper_1._client_manager._client,
            sparqlwrapper_2._client_manager.client,
            sparqlwrapper_2._client_manager._client,
        ]
    )

    assert all(
        not _client.is_closed
        for _client in [
            sparqlwrapper_1._client_manager.client,
            sparqlwrapper_1._client_manager._client,
            sparqlwrapper_2._client_manager.client,
            sparqlwrapper_2._client_manager._client,
        ]
    )

    client.close()

    assert all(
        _client.is_closed
        for _client in [
            sparqlwrapper_1._client_manager.client,
            sparqlwrapper_1._client_manager._client,
            sparqlwrapper_2._client_manager.client,
            sparqlwrapper_2._client_manager._client,
        ]
    )


def test_managed_context_client():
    """Check managed client status is a SPARQLOperationWrapper context."""

    sparqlwrapper = SPARQLWrapper(sparql_endpoint="https://some.endoint")
    assert sparqlwrapper._client_manager._client is None

    with sparqlwrapper as wrapper_context:
        assert sparqlwrapper is wrapper_context
        assert isinstance(wrapper_context._client_manager.client, httpx.Client)

        context_client = wrapper_context._client_manager._client
        context_client_property = wrapper_context._client_manager.client

        assert context_client is context_client_property

    assert context_client.is_closed
    assert context_client_property.is_closed


def test_shared_context_client():
    """Check shard client status is a SPARQLOperationWrapper context."""
    client = httpx.Client()

    sparqlwrapper = SPARQLWrapper(sparql_endpoint="https://some.endoint", client=client)
    assert sparqlwrapper._client_manager._client is client

    with sparqlwrapper as wrapper_context:
        assert sparqlwrapper is wrapper_context

        context_client = wrapper_context._client_manager._client
        context_client_property = wrapper_context._client_manager.client

        assert context_client is context_client_property is client

        assert all(
            not _client.is_closed
            for _client in [client, context_client, context_client_property]
        )

    assert all(
        _client.is_closed
        for _client in [client, context_client, context_client_property]
    )



from collections.abc import Iterator
import json

import httpx
from rdflib import BNode, Graph, Literal, URIRef
from sparqlx.types import SPARQLResultBinding, SPARQLResultBindingValue


def _convert_bindings(
    response: httpx.Response,
) -> list[SPARQLResultBinding]:
    """Get flat dicts from a SPARQL SELECT JSON response."""

    try:
        json_response = response.json()
    except (
        json.JSONDecodeError
    ) as error:  # pragma: no cover ; this should be unreachable
        error.add_note("Note that convert=True requires JSON as response format.")
        raise error

    variables = json_response["head"]["vars"]
    response_bindings = json_response["results"]["bindings"]

    def _get_binding_pairs(binding) -> Iterator[tuple[str, SPARQLResultBindingValue]]:
        """Generate key value pairs from response_bindings.

        The 'type' and 'datatype' fields of the JSON response
        are examined to cast values to Python types according to RDFLib.
        """
        for var in variables:
            if (binding_data := binding.get(var, None)) is None:
                yield (var, None)
                continue

            match binding_data["type"]:
                case "uri":
                    yield (var, URIRef(binding_data["value"]))
                case "literal":
                    literal = Literal(
                        binding_data["value"],
                        datatype=binding_data.get("datatype", None),
                    )

                    literal_to_python = literal.toPython()
                    yield (var, literal_to_python)

                case "bnode":
                    yield (var, BNode(binding_data["value"]))
                case _:  # pragma: no cover
                    assert False, "This should never happen."

    return [dict(_get_binding_pairs(binding)) for binding in response_bindings]


def _convert_graph(response: httpx.Response) -> Graph:
    """Convert the content of an HTTP response to an rdflib.Graph.

    Note: httpx.Response.headers is always an instance of httpx.Headers
    (a mutable mapping).
    """
    _format: str | None = (
        content_type
        if (content_type := response.headers.get("content-type")) is None
        else content_type.split(";")[0].strip()
    )

    graph = Graph().parse(response.content, format=_format)
    return graph


def _convert_ask(response: httpx.Response) -> bool:
    return response.json()["boolean"]



"""Basic tests for the SPARQLWrapper.from_rdf_source constructor."""

from io import StringIO

import pytest
from rdflib import Dataset, Graph, URIRef
from sparqlx import SPARQLWrapper
from sparqlx.types import SPARQLResultBinding, SelectQuery

from utils import sparql_result_set_equal


@pytest.mark.parametrize("graph_source", [Graph(), Dataset()])
def test_sparqlwrapper_from_graph_source(graph_source):
    """Check that Graph objects passed to the constructor get set as endpoints."""
    wrapper = SPARQLWrapper.from_rdf_source(rdf_source=graph_source)

    assert wrapper._sparql_endpoint.graph is graph_source
    assert wrapper._update_endpoint.graph is graph_source


@pytest.mark.parametrize(
    "rdf_source",
    [
        StringIO("<urn:s> <urn:p> <urn:o>."),
        Graph().parse(StringIO("<urn:s> <urn:p> <urn:o>.")),
        Dataset().parse(StringIO("<urn:s> <urn:p> <urn:o>.")),
    ],
)
def test_sparqlwrapper_from_source(rdf_source):
    """Construct a SPARQLWrapper from an RDF source and check a query result."""

    wrapper = SPARQLWrapper.from_rdf_source(rdf_source=rdf_source)

    result = wrapper.query(
        SelectQuery("select ?s ?p ?o where {?s ?p ?o}"), convert=True
    )
    excepted: list[SPARQLResultBinding] = [
        {
            "s": URIRef("urn:s"),
            "p": URIRef("urn:p"),
            "o": URIRef("urn:o"),
        }
    ]

    assert sparql_result_set_equal(result, excepted)



from typing import NamedTuple

import pytest

from conftest import RDFLibGraphEndpoints
from sparqlx import SPARQLWrapper
from utils import parse_response_qs


class ProtocolRequestParameters(NamedTuple):
    kwargs: dict
    expected: dict


query_request_params = [
    ProtocolRequestParameters(
        kwargs={"named_graph_uri": "https://named.graph"},
        expected={
            "query": ["select * where {?s ?p ?o}"],
            "named-graph-uri": ["https://named.graph"],
        },
    ),
    ProtocolRequestParameters(
        kwargs={
            "default_graph_uri": "https://default.graph",
            "named_graph_uri": "https://named.graph",
        },
        expected={
            "query": ["select * where {?s ?p ?o}"],
            "default-graph-uri": ["https://default.graph"],
            "named-graph-uri": ["https://named.graph"],
        },
    ),
    ProtocolRequestParameters(
        kwargs={
            "default_graph_uri": "https://default.graph",
            "named_graph_uri": "https://named.graph",
            "version": "1.2",
        },
        expected={
            "query": ["select * where {?s ?p ?o}"],
            "default-graph-uri": ["https://default.graph"],
            "named-graph-uri": ["https://named.graph"],
            "version": ["1.2"],
        },
    ),
    ProtocolRequestParameters(
        kwargs={
            "default_graph_uri": "https://default.graph",
            "named_graph_uri": ["https://named.graph", "https://othernamed.graph"],
        },
        expected={
            "query": ["select * where {?s ?p ?o}"],
            "default-graph-uri": ["https://default.graph"],
            "named-graph-uri": ["https://named.graph", "https://othernamed.graph"],
        },
    ),
    ProtocolRequestParameters(
        kwargs={
            "default_graph_uri": [
                "https://default.graph",
                "https://otherdefault.graph",
            ],
            "named_graph_uri": ["https://named.graph", "https://othernamed.graph"],
        },
        expected={
            "query": ["select * where {?s ?p ?o}"],
            "default-graph-uri": [
                "https://default.graph",
                "https://otherdefault.graph",
            ],
            "named-graph-uri": ["https://named.graph", "https://othernamed.graph"],
        },
    ),
]

update_request_params = [
    ProtocolRequestParameters(
        kwargs={"using_graph_uri": "https://named.graph"},
        expected={
            "update": ["insert data {}"],
            "using-graph-uri": ["https://named.graph"],
        },
    ),
    ProtocolRequestParameters(
        kwargs={"using_graph_uri": ["https://named.graph", "https://othernamed.graph"]},
        expected={
            "update": ["insert data {}"],
            "using-graph-uri": ["https://named.graph", "https://othernamed.graph"],
        },
    ),
    ProtocolRequestParameters(
        kwargs={
            "using_named_graph_uri": [
                "https://named.graph",
                "https://othernamed.graph",
            ],
        },
        expected={
            "update": ["insert data {}"],
            "using-named-graph-uri": [
                "https://named.graph",
                "https://othernamed.graph",
            ],
        },
    ),
    ProtocolRequestParameters(
        kwargs={
            "using_graph_uri": "https://named.graph",
            "using_named_graph_uri": "https://othernamed.graph",
        },
        expected={
            "update": ["insert data {}"],
            "using-graph-uri": ["https://named.graph"],
            "using-named-graph-uri": ["https://othernamed.graph"],
        },
    ),
    ProtocolRequestParameters(
        kwargs={
            "using_graph_uri": ["https://named.graph"],
            "using_named_graph_uri": ["https://othernamed.graph"],
        },
        expected={
            "update": ["insert data {}"],
            "using-graph-uri": ["https://named.graph"],
            "using-named-graph-uri": ["https://othernamed.graph"],
        },
    ),
]


@pytest.mark.parametrize("param", query_request_params)
def test_sparqlwrapper_query_request_params(param, triplestore):
    if isinstance(triplestore, RDFLibGraphEndpoints):
        pytest.skip(reason="SPARQL Protocol parameters currently not implemented.")

    sparqlwrapper = SPARQLWrapper(sparql_endpoint=triplestore.sparql_endpoint)
    response = sparqlwrapper.query("select * where {?s ?p ?o}", **param.kwargs)

    assert parse_response_qs(response) == param.expected


@pytest.mark.parametrize("param", update_request_params)
def test_sparqlwrapper_update_request_params(param, triplestore):
    if isinstance(triplestore, RDFLibGraphEndpoints):
        pytest.skip(reason="SPARQL Protocol parameters currently not implemented.")

    sparqlwrapper = SPARQLWrapper(update_endpoint=triplestore.update_endpoint)
    response = sparqlwrapper.update("insert data {}", **param.kwargs)

    assert parse_response_qs(response) == param.expected



"""Pytest entry point for sad paths tests of SPARQLWrapper with rdflib.Graph targets."""

import httpx
import pytest
from rdflib import Dataset, Graph
from sparqlx import SPARQLWrapper

from utils import acall


@pytest.mark.parametrize("graph", [Graph(), Dataset()])
@pytest.mark.parametrize("query_method", ["query", "aquery"])
@pytest.mark.parametrize("update_method", ["update", "aupdate"])
@pytest.mark.parametrize(
    "query_params",
    [
        {"default_graph_uri": ""},
        {"named_graph_uri": ""},
        {"default_graph_uri": "", "named_graph_uri": ""},
    ],
)
@pytest.mark.parametrize(
    "update_params",
    [
        {"using_graph_uri": ""},
        {"using_named_graph_uri": ""},
        {"using_graph_uri": "", "using_named_graph_uri": ""},
    ],
)
@pytest.mark.asyncio
async def test_sparqlwrapper_graph_target_uriparameters(
    graph, query_params, update_params, query_method, update_method
):
    """Check that URI parameters on rdflib.Graph targets raise NotImplementedError."""

    sparqlwrapper = SPARQLWrapper(sparql_endpoint=graph, update_endpoint=graph)

    partial_fail_msg = "are currently not supported for rdflib.Graph targets."

    with pytest.raises(NotImplementedError, match=partial_fail_msg):
        await acall(
            sparqlwrapper,
            query_method,
            query="select * where {?s ?p ?o}",
            **query_params,
        )

    with pytest.raises(NotImplementedError, match=partial_fail_msg):
        await acall(
            sparqlwrapper,
            update_method,
            update_request="insert data {<urn:s> <urn:p> <urn:o>}",
            **update_params,
        )


@pytest.mark.parametrize("graph", [Graph(), Dataset()])
@pytest.mark.asyncio
async def test_sparqlwrapper_graph_target_streaming(graph):
    """Check that response streaming with rdflib.Graph targets raises NotImplementedError."""

    sparqlwrapper = SPARQLWrapper(sparql_endpoint=graph, update_endpoint=graph)

    fail_msg = "Response streaming is currently not supported for rdflib.Graph targets."
    query = "select * where {?s ?p ?o}"

    stream = sparqlwrapper.query_stream(query, chunk_size=1)
    astream = sparqlwrapper.aquery_stream(query, chunk_size=1)

    with pytest.raises(NotImplementedError, match=fail_msg):
        [chunk for chunk in stream]

    with pytest.raises(NotImplementedError, match=fail_msg):
        [chunk async for chunk in astream]


def test_sparqlwrapper_graph_target_http_500():
    """Cause an HTTP 500 with an rdflib.Graph target."""

    graph = Graph()
    sparqlwrapper = SPARQLWrapper(sparql_endpoint=graph, update_endpoint=graph)

    partial_fail_msg_500 = "Server error '500 Internal Server Error'"
    partial_fail_msg_400 = "Client error '400 Bad Request'"

    with pytest.raises(httpx.HTTPStatusError, match=partial_fail_msg_500):
        sparqlwrapper.query("select * where {graph ?g {?s ?p ?o}}")

    with pytest.raises(httpx.HTTPStatusError, match=partial_fail_msg_500):
        sparqlwrapper.update(
            "insert {graph ?g {<urn:s> <urn:p> <urn:o>}} where {graph ?g {?s ?p ?o}}"
        )

    with pytest.raises(httpx.HTTPStatusError, match=partial_fail_msg_400):
        sparqlwrapper.update("invalid SPARQL")



from sparqlx.sparqlwrapper import SPARQLWrapper
from sparqlx.types import (
    AskQuery,
    ConstructQuery,
    DescribeQuery,
    SPARQLQuery,
    SPARQLQueryType,
    SPARQLQueryTypeLiteral,
    SelectQuery,
)
from sparqlx.utils.utils import SPARQLParseException

__all__ = (
    "SPARQLWrapper",
    "SPARQLParseException",
    "AskQuery",
    "ConstructQuery",
    "DescribeQuery",
    "SPARQLQuery",
    "SPARQLQueryType",
    "SelectQuery",
    "SPARQLQueryTypeLiteral",
)



"""Custom httpx.Transports for rdflb.Graph targets."""

import asyncio

import httpx
import pyparsing
from rdflib import Graph
from sparqlx.types import RequestDataValue, SPARQLQuery


class _RDFLibQueryTransportBase:
    def __init__(
        self,
        query: SPARQLQuery,
        graph: Graph,
        version: str | None = None,
        default_graph_uri: RequestDataValue = None,
        named_graph_uri: RequestDataValue = None,
    ) -> None:
        if any(param is not None for param in [default_graph_uri, named_graph_uri]):
            msg = (
                "URI Parameters 'default-graph-uri' and 'named-graph-uri' "
                "are currently not supported for rdflib.Graph targets."
            )

            raise NotImplementedError(msg)

        self._query = query
        self._graph = graph

    def _handle_request(self, request: httpx.Request) -> httpx.Response:
        _format = request.headers["Accept"]

        try:
            content = self._graph.query(self._query).serialize(format=_format)
        except (
            pyparsing.exceptions.ParseBaseException
        ) as e:  # pragma: no cover ; this is currently unreachable
            return httpx.Response(
                status_code=400,
                headers={
                    "Content-Type": "text/plain; charset=utf-8",
                    "Server": "rdflib.Graph",
                },
                request=request,
                content=str(e),
            )
        except Exception as e:
            return httpx.Response(
                status_code=500,
                headers={
                    "Content-Type": "text/plain; charset=utf-8",
                    "Server": "rdflib.Graph",
                },
                request=request,
                content=str(e),
            )

        else:
            return httpx.Response(
                status_code=200,
                headers={
                    "Content-Type": _format,
                    "Server": "rdflib.Graph",
                },
                request=request,
                content=content,
            )


class RDFLibQueryTransport(_RDFLibQueryTransportBase, httpx.BaseTransport):
    def handle_request(self, request: httpx.Request) -> httpx.Response:
        return self._handle_request(request)


class AsyncRDFLibQueryTransport(_RDFLibQueryTransportBase, httpx.AsyncBaseTransport):
    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        return await asyncio.to_thread(self._handle_request, request)


class _RDFLibUpdateTransportBase(httpx.BaseTransport):
    def __init__(
        self,
        update_request: str,
        graph: Graph,
        version: str | None = None,
        using_graph_uri: RequestDataValue = None,
        using_named_graph_uri: RequestDataValue = None,
    ) -> None:
        if any(param is not None for param in [using_graph_uri, using_named_graph_uri]):
            msg = (
                "URI Parameters 'using-graph-uri' and 'using-named-graph-uri' "
                "are currently not supported for rdflib.Graph targets."
            )

            raise NotImplementedError(msg)

        self._update_request = update_request
        self._graph = graph

    def _handle_request(self, request: httpx.Request) -> httpx.Response:
        try:
            self._graph.update(self._update_request)
        except pyparsing.exceptions.ParseBaseException as e:
            return httpx.Response(
                status_code=400,
                headers={
                    "Content-Type": "text/plain; charset=utf-8",
                    "Server": "rdflib.Graph",
                },
                request=request,
                content=str(e),
            )
        except Exception as e:
            return httpx.Response(
                status_code=500,
                headers={
                    "Content-Type": "text/plain; charset=utf-8",
                    "Server": "rdflib.Graph",
                },
                request=request,
                content=str(e),
            )

        else:
            return httpx.Response(
                status_code=200,
                headers={
                    "Content-Length": "0",
                    "Content-Type": "text/plain; charset=utf-8",
                    "Server": "rdflib.Graph",
                },
                request=request,
                content=b"",
            )


class RDFLibUpdateTransport(_RDFLibUpdateTransportBase, httpx.BaseTransport):
    def handle_request(self, request: httpx.Request) -> httpx.Response:
        return self._handle_request(request)


class AsyncRDFLibUpdateTransport(_RDFLibUpdateTransportBase, httpx.AsyncBaseTransport):
    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        return await asyncio.to_thread(self._handle_request, request)



"""Pytest entry point for basic SPARQLWrapper Query Operation tests."""

import asyncio
from collections.abc import Callable
import datetime
from decimal import Decimal
import operator
from typing import NamedTuple
from typing import Any
import warnings

import httpx
import pytest
from rdflib import BNode, Graph, Literal, URIRef, XSD
from rdflib.compare import isomorphic
from sparqlx import SPARQLWrapper
from sparqlx.utils.operation_parameters import (
    rdf_response_format_map,
    sparql_result_response_format_map,
)

from conftest import FusekiEndpoints, RDFLibGraphEndpoints
from data.queries import (
    ask_query_false,
    ask_query_true,
    construct_query_x_values,
    describe_query,
    select_query_bnode,
    select_query_types,
    select_query_xy_values,
)
from utils import acall


class QueryOperationParameter(NamedTuple):
    query: str
    expected: bool | list[dict] | Graph
    compare: Callable[[Any, Any], bool] = operator.eq


ntriples_data = """
<urn:s> <urn:p> "1"^^<http://www.w3.org/2001/XMLSchema#integer> .
<urn:s> <urn:p> "3"^^<http://www.w3.org/2001/XMLSchema#integer> .
<urn:s> <urn:p> "2"^^<http://www.w3.org/2001/XMLSchema#integer> .
"""
expected_graph = Graph().parse(data=ntriples_data)


params = [
    QueryOperationParameter(
        query=select_query_xy_values,
        expected=[{"x": 1, "y": 2}, {"x": 3, "y": 4}],
    ),
    QueryOperationParameter(query=ask_query_true, expected=True),
    QueryOperationParameter(query=ask_query_false, expected=False),
    QueryOperationParameter(
        query=construct_query_x_values, expected=expected_graph, compare=isomorphic
    ),
    #
    QueryOperationParameter(
        query=select_query_types,
        expected=[
            {"x": 2},
            {"x": Decimal("2.2")},
            {"x": None},
            {"x": URIRef("https://test.uri")},
            {"x": datetime.date(2024, 1, 1)},
            {"x": Literal("2024", datatype=XSD.gYear)},
            {"x": Literal("2024-01", datatype=XSD.gYearMonth)},
        ],
    ),
    QueryOperationParameter(
        query=select_query_bnode,
        expected=[{"x": BNode()}],
        compare=lambda x, y: all(
            isinstance(v, BNode) for v in map(lambda r: list(r)[0]["x"], [x, y])
        ),
    ),
    QueryOperationParameter(query=describe_query, expected=Graph(), compare=isomorphic),
]


@pytest.mark.parametrize("method", ["query", "aquery"])
@pytest.mark.parametrize("query_method", ["GET", "POST", "POST-direct"])
@pytest.mark.parametrize("param", params)
@pytest.mark.asyncio
async def test_sparqlwrapper_query(method, query_method, param, triplestore):
    """Run a query with convert=True and compare the result to the expected bindings."""

    if isinstance(triplestore, RDFLibGraphEndpoints) and param.query == ask_query_false:
        pytest.skip(reason="Bug in RDFLib, ask_query_false returns True.")

    endpoint = triplestore.sparql_endpoint
    sparqlwrapper = SPARQLWrapper(sparql_endpoint=endpoint, query_method=query_method)

    result_converted = await acall(
        sparqlwrapper, method, query=param.query, convert=True
    )

    assert param.compare(result_converted, param.expected)


#### ask queries and tsv failing for rdflib.Graph
@pytest.mark.parametrize("method", ["query", "aquery"])
@pytest.mark.parametrize("query_method", ["GET", "POST", "POST-direct"])
@pytest.mark.parametrize(
    "query",
    [
        select_query_xy_values,
        ask_query_true,
        ask_query_false,
        select_query_bnode,
        select_query_types,
    ],
)
@pytest.mark.parametrize(
    "response_format",
    [None, *sparql_result_response_format_map.keys()],
)
@pytest.mark.asyncio
async def test_sparqlwrapper_query_binding_result_formats(
    method,
    query_method,
    query,
    response_format,
    triplestore,
):
    """Run SELECT and ASK queries with bindings result formats."""

    if isinstance(triplestore, RDFLibGraphEndpoints):
        if query == ask_query_false:
            pytest.skip(reason="Bug in RDFLib, ask_query_false returns True.")
        if response_format in ["csv", "tsv"]:
            pytest.skip(reason="CSV/TSV is not well supported in rdflib.Graph.query.")

    endpoint = triplestore.sparql_endpoint
    sparqlwrapper = SPARQLWrapper(sparql_endpoint=endpoint, query_method=query_method)

    result = await acall(
        sparqlwrapper, method, query=query, response_format=response_format
    )

    assert result.content


@pytest.mark.parametrize("method", ["query", "aquery"])
@pytest.mark.parametrize("query_method", ["GET", "POST", "POST-direct"])
@pytest.mark.parametrize("query", [construct_query_x_values, describe_query])
@pytest.mark.parametrize(
    "response_format", [None, *rdf_response_format_map.keys(), "application/n-triples"]
)
@pytest.mark.asyncio
async def test_sparqlwrapper_query_graph_result_formats(
    method,
    query_method,
    query,
    response_format,
    triplestore_with_data,
):
    """Run CONSTRUCT and DESCRIBE queries with graph result formats.

    The tests uses the oxigraph_service_with_data fixture in order
    to retrieve a non-empty graph object on DESCRIBE queries.

    Note: Fuseki responds with 406 if 'application/rdf+xml' is requested as graph format;
    the store can handle 'text/xml' though, so this can be considered a bug in Fuseki.
    """
    if isinstance(triplestore_with_data, FusekiEndpoints) and response_format == "xml":
        pytest.skip(
            reason="Fuseki currently only handles 'text/xml', not 'application/rdf+xml."
        )

    endpoint = triplestore_with_data.sparql_endpoint
    sparqlwrapper = SPARQLWrapper(sparql_endpoint=endpoint, query_method=query_method)

    result = await acall(
        sparqlwrapper, method, query=query, response_format=response_format
    )

    result_converted = await acall(
        sparqlwrapper,
        method,
        query=query,
        convert=True,
        response_format=response_format,
    )

    assert result
    assert result_converted


@pytest.mark.parametrize("query_method", ["GET", "POST", "POST-direct"])
@pytest.mark.asyncio
async def test_sparqlwrapper_warn_open_client(triplestore, query_method):
    """Pass a client/aclient and check if a warning is emitted."""
    if isinstance(triplestore, RDFLibGraphEndpoints):
        pytest.skip(reason="Client sharing is not supported for rdflib.Graph targets.")

    endpoint = triplestore.sparql_endpoint

    client = httpx.Client()
    aclient = httpx.AsyncClient()

    sparqlwrapper = SPARQLWrapper(
        sparql_endpoint=endpoint,
        client=client,
        aclient=aclient,
        query_method=query_method,
    )

    def _get_msg(client):
        return (
            f"httpx Client instance '{client}' is not managed. "
            "Client.close/AsyncClient.aclose should be called at some point."
        )

    with pytest.warns(UserWarning, match=_get_msg(client)):
        sparqlwrapper.query(select_query_xy_values)

    with pytest.warns(UserWarning, match=_get_msg(aclient)):
        await sparqlwrapper.aquery(select_query_xy_values)


@pytest.mark.parametrize("query_method", ["GET", "POST", "POST-direct"])
@pytest.mark.parametrize(
    "query",
    [
        select_query_xy_values,
        ask_query_false,
        ask_query_true,
        construct_query_x_values,
        describe_query,
    ],
)
@pytest.mark.asyncio
async def test_sparql_wrapper_context_managers(query, query_method, triplestore):
    """Check unmanaged client/aclient status in context managers.

    The tests assert that client/aclient is open within a context
    and is closed after leaving the context."""
    endpoint = triplestore.sparql_endpoint

    client = httpx.Client()
    aclient = httpx.AsyncClient()

    sparqlwrapper = SPARQLWrapper(
        sparql_endpoint=endpoint,
        client=client,
        aclient=aclient,
        query_method=query_method,
    )

    with sparqlwrapper as context_wrapper:
        result_1 = context_wrapper.query(query=query)
        assert not client.is_closed

    async with sparqlwrapper as context_wrapper:
        result_2 = await context_wrapper.aquery(query=query)
        assert not aclient.is_closed

    assert client.is_closed
    assert aclient.is_closed

    assert result_1.content == result_2.content


@pytest.mark.parametrize("query_method", ["GET", "POST", "POST-direct"])
@pytest.mark.parametrize(
    "query",
    [
        select_query_xy_values,
        ask_query_false,
        ask_query_true,
        construct_query_x_values,
        describe_query,
        select_query_types,
        # note: bnode query cannot be compared
    ],
)
@pytest.mark.asyncio
async def test_sparqlwrapper_streaming(query_method, query, triplestore):
    """Sync/async stream a response and compare lists of chunks."""

    if isinstance(triplestore, RDFLibGraphEndpoints):
        pytest.skip(
            reason="Response streaming is currently not supported for rdflib.Graph targets."
        )

    endpoint = triplestore.sparql_endpoint
    sparqlwrapper = SPARQLWrapper(sparql_endpoint=endpoint, query_method=query_method)

    stream = sparqlwrapper.query_stream(query, chunk_size=1)
    astream = sparqlwrapper.aquery_stream(query, chunk_size=1)

    chunks = [chunk for chunk in stream]
    achunks = [chunk async for chunk in astream]

    assert chunks == achunks


@pytest.mark.parametrize("query_method", ["GET", "POST", "POST-direct"])
@pytest.mark.parametrize(
    "query",
    [
        select_query_xy_values,
        ask_query_false,
        ask_query_true,
        construct_query_x_values,
        describe_query,
    ],
)
def test_sparqlwrapper_queries(query_method, query, triplestore):
    """Compare the results of SPARQLWrapper.queries and manually running SPARQLWrapper.aquery.

    The test also checks that no client manager warnings are emitted
    from SPARQLWrapper.queries. This is achieved by locally setting
    the warnings filter to raise an error for all warnings;
    i.e. if no warning is emitted, the test passes because no exceptions
    are raised from any warning.
    """

    endpoint = triplestore.sparql_endpoint
    sparqlwrapper = SPARQLWrapper(sparql_endpoint=endpoint, query_method=query_method)

    queries: list[str] = [query for _ in range(5)]

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        results_queries = sparqlwrapper.queries(*queries)

    async def _runner():
        async with asyncio.TaskGroup() as tg:
            tasks = [tg.create_task(sparqlwrapper.aquery(query)) for query in queries]

        return map(asyncio.Task.result, tasks)

    results_aqueries = asyncio.run(_runner())

    assert all(
        response_1.content == response_2.content
        for response_1, response_2 in zip(results_queries, results_aqueries)
    )
